SystemOrganization addCategory: #'RPackage-Core'!

!ClassDescription methodsFor: '*rpackage-core' stamp: 'sd 10/29/2009 20:31'!
compileSilently: code 
	"Compile the code and classify the resulting method in the given category, leaving no trail in the system log, nor in any change set, nor in the 'recent submissions' list. This should only be used when you know for sure that the compilation will succeed."

	^ self compileSilently: code classified: 'not defined category' notifying: nil.! !

!ClassDescription methodsFor: '*rpackage-core' stamp: 'tg 6/13/2010 17:32'!
extendingPackages
	"the extending packages of a class are the packages that extend it."
	^ RPackage organizer extendingPackagesOf: self
	! !

!ClassDescription methodsFor: '*rpackage-core' stamp: 'StephaneDucasse 4/20/2010 22:10'!
isDefinedInPackage: aPackage
	"returns true if aPackage contains the definitino of this class"
	
	^ aPackage includesClass: self.
	! !

!ClassDescription methodsFor: '*rpackage-core' stamp: 'StephaneDucasse 5/15/2010 19:54'!
isExtended

	^ self extendingPackages isEmpty! !

!ClassDescription methodsFor: '*rpackage-core' stamp: 'tg 3/16/2010 17:53'!
isExtendedInPackage: aPackage
	"returns true if aPackage defines an extension to this class"
	^ aPackage extendsClass: self.
	! !

!ClassDescription methodsFor: '*rpackage-core' stamp: 'tg 6/13/2010 17:19'!
package
	^ RPackage organizer packageOf: self.	! !

!ClassDescription methodsFor: '*rpackage-core' stamp: 'tg 3/16/2010 17:36'!
packageFromOrganizer: anOrganizer
	"returns the package that defines this class"
	^ anOrganizer packageOf: self.
	! !

!ClassDescription methodsFor: '*rpackage-core' stamp: 'sd 5/19/2010 16:31'!
packages
	"the extending packages of a class are the packages that extend it."
	
	^  self extendingPackages asSet copy add: self package; yourself! !

!CompiledMethod methodsFor: '*rpackage-core' stamp: 'tg 3/13/2010 15:13'!
isDefinedInPackage: anRPackage
	^ anRPackage includesDefinedSelector: self selector ofClass: self methodClass! !

!CompiledMethod methodsFor: '*rpackage-core' stamp: 'tg 3/13/2010 13:50'!
isExtensionInPackage: anRPackage
	^ anRPackage includesExtensionSelector: self selector ofClass: self methodClass! !

!CompiledMethod methodsFor: '*rpackage-core' stamp: 'tg 3/13/2010 16:37'!
packageFromOrganizer: anRPackageOrganizer
	^ anRPackageOrganizer packageDefiningOrExtendingMethod: self! !

!SystemOrganizer methodsFor: '*RPackage-Core' stamp: 'cyrille.delaunay 11/24/2010 13:11'!
classify: aName under: categorySymbol

	"I overrid this method to emit the right event when we recategorized a class. We can use this method to move a class from its original category to a new category. If it's the case we should emit the event 'class rectegorized' to be sure that all object listening to this event are alerted"
	|oldCategory|
	oldCategory := self elementCategoryDict at: aName asSymbol ifAbsent: [
		"if we are not able to find the class. it means that it not have been yet categorized"
		^ super classify: aName under: categorySymbol.
		].
	super classify: aName under: categorySymbol.
	"SystemChangeNotifier uniqueInstance class: (Smalltalk globals at: aName asSymbol) 	recategorizedFrom: oldCategory to: categorySymbol."! !

Object subclass: #RPackage
	instanceVariableNames: 'classDefinedSelectors metaclassDefinedSelectors classExtensionSelectors metaclassExtensionSelectors classes name'
	classVariableNames: 'PackageGlobalOrganizer'
	poolDictionaries: ''
	category: 'RPackage-Core'!

!RPackage commentStamp: 'sd 4/25/2010 17:49' prior: 0!
A RPackage2 is a new implementation of package. Note that I does not touch classes but I'm just a layer on top of classes therefore removing a method from a package does not change the underlying class. 


Instance Variables
	classDefinedSelectors:		Dictionary
	classExtensionSelectors:		Dictionary
	metaclassDefinedSelectors:		Dictionary
	metaclassExtensionSelectors:		Dictionary
	packageName:		String
	classes: OrderedCollection of Symbols

The reason we use four dictionaries is that this way we only store #Point in both as the class and the metaclass are involved. We do not have to with Point_class problem when comparing defined and extended classes. The idea is taken from the RBSelectorEnvironment. 					
		
Class Invariant one:
	Classes should not contain metaclass name, but only class names
	
	

Implementation notes 
====================
This class went over 3 internal representations implementation. 
	- first: 	a list of class + two dictionaries: class * methods
	This was not good since we add to go all the time over the list of classes.
	- second: 4 dictionaries class * selectors 
	This was not good since we want to have defined classes without methods. 
	- third: 4 dictionaries + definedClasses

Originally I wanted to avoid to have a defined class list and I wanted to avoid to have to declare the class as defined. But this is not really good since
	- we will want to know if a class definition (without method for example is defined in a given package)
	- second this is easier to hook the packageOf: behavior (we register the package class when the class is 	added to the package).

defined classes information is redundant with the dictionary keys of defined methods but we would have to check and register the class to the packageOrganizer the first time a method is defined then also check on remove to unregister the class from the packageOrganizer.

Adding a method does not define the class as a defined package class. This has to be done explictly. The reason for this choice is that a class should register to the packageOrganizer and that I do not want to do it each time a method is added and I do not want to test it each time a method is added. Now this is done only when the class is declared as defined. 
We could also give the complete freedom to the client to register the class but I thought it was a good compromise. 
	
Todo
====
	Next: 
	- should use resources for the completeSetupTest
	- register to systemNotifier
	- build up a synchronizer that import PackageInfo.
	We should port all the tests.
	We should test packages and extendingPackages defined on classDescription
	We should add MC support.
	- add events.

	

	
!

!RPackage class methodsFor: 'initialize-release' stamp: 'TudorGirba 11/25/2010 23:53'!
initialize
	super initialize.
	self organizer: RPackageOrganizer default.
	self organizer initializeFromSystem! !

!RPackage class methodsFor: 'instance creation' stamp: 'tg 6/13/2010 17:22'!
named: aString

	| newPackage |
	newPackage := self new name: aString ;yourself.
	self organizer registerPackage: newPackage.
	^ newPackage! !

!RPackage class methodsFor: 'protected' stamp: 'tg 6/13/2010 16:59'!
organizer
	^ PackageGlobalOrganizer! !

!RPackage class methodsFor: 'protected' stamp: 'cyrille.delaunay 10/29/2010 14:31'!
organizer: aPackageOrganizer

	PackageGlobalOrganizer := aPackageOrganizer! !

!RPackage class methodsFor: 'protected' stamp: 'cyrille.delaunay 11/3/2010 18:33'!
withOrganizer: aNewOrganizer do: aBlock
	"Perform an action locally to aNewOrganizer. Does not impact any other organizers."
	
	| old |
	[ old := self organizer.
	self organizer: aNewOrganizer.
	aBlock cull: aNewOrganizer.] ensure: [ self organizer: old ] ! !

!RPackage methodsFor: 'class addition' stamp: 'DamienPollet 10/1/2010 11:37'!
addClassDefinition: aClass
	"Add only the class definition of aClass, as a class defined by the receiver. The term classdefinition denotes the fact that we do not treat its methods. aClass methods are not added to the package. aClass can be either a class or a metaclass"
	
	self addClassDefinitionName: aClass theNonMetaClass name.
	! !

!RPackage methodsFor: 'class addition' stamp: 'cyrille.delaunay 11/25/2010 11:23'!
addClassDefinitionName: aClassName
	"aClassName should only be a class (and not a metaclass name)"
	
	| aClassNameSymbol |
	('* class' match: aClassName) ifTrue: [^ self error: 'no metaclass name'].
	aClassNameSymbol := aClassName asSymbol.
	(classes includes: aClassNameSymbol)
		ifFalse: 
			[classes add: aClassNameSymbol.
			self extendedMethodsShouldBecomeDefinedWhenAddingClassName: aClassNameSymbol.
			self registerClassName: aClassNameSymbol
			]
	! !

!RPackage methodsFor: 'compiled method handling' stamp: 'cyrille.delaunay 10/27/2010 15:38'!
addMethod: aCompiledMethod 
	"Add the method to the receiver as a defined method if the class is  defined in it, else as an extension."

	| mthClass |
	mthClass := aCompiledMethod methodClass. 
	(self includesClass: mthClass )
		ifFalse: [mthClass isMeta
					ifTrue: [(metaclassExtensionSelectors 
								at: mthClass theNonMetaClass name  
								ifAbsentPut: [Set new]) add: aCompiledMethod selector.]
					ifFalse: [(classExtensionSelectors 
								at: mthClass name 
								ifAbsentPut: [Set new]) add: aCompiledMethod selector].
				"we added a method extension so the receiver is an extending package of the class"
				self organizer registerExtendingPackage: self forClass: mthClass]
		ifTrue: [mthClass isMeta
					ifTrue: [(metaclassDefinedSelectors 
								at: mthClass theNonMetaClass name  
								ifAbsentPut: [Set new]) add: aCompiledMethod selector]
					ifFalse: [(classDefinedSelectors 
								at: mthClass name 
								ifAbsentPut: [Set new]) add: aCompiledMethod selector]].	
	^ aCompiledMethod.! !

!RPackage methodsFor: 'selector handling' stamp: 'StephaneDucasse 4/20/2010 22:56'!
addSelector: aSelector ofClassName: aClassName
	"Add the method to the receiver. If the class is not locally defined in that package then the method is defined as a method extension: ie extending another package. Note that this method does not add the method to the class, it just records in the package that the method is packaged."
	
	( self includesClassNamed: aClassName)
		ifFalse: [(classExtensionSelectors at: aClassName asSymbol ifAbsentPut: [Set new]) add: aSelector]
		ifTrue: [(classDefinedSelectors at: aClassName asSymbol ifAbsentPut: [Set new]) add: aSelector].
	^ aSelector.! !

!RPackage methodsFor: 'selector handling' stamp: 'StephaneDucasse 4/20/2010 22:56'!
addSelector: aSelector ofMetaclassName: aClassName
	"Add the method to the receiver. If the class is not locally defined in that package then the method is defined as a method extension: ie extending another package. Note that this method does not add the method to the class, it just records in the package that the method is packaged. aClassName is the sole instance class name and not its metaclass one: i.e. adding Point class>>new is done as addSelector: #new ofMetaclassName: #Point"
	
	( self includesClassNamed: aClassName) 
		ifFalse: [(metaclassExtensionSelectors at: aClassName asSymbol ifAbsentPut: [Set new]) add: aSelector]
		ifTrue: [(metaclassDefinedSelectors at: aClassName asSymbol ifAbsentPut: [Set new]) add: aSelector].
	^ aSelector.! !

!RPackage methodsFor: 'slices' stamp: 'StephaneDucasse 3/11/2010 18:46'!
classDefinedSlicesDo: aBlock
	"This method iterates over the defined class and their associated selectors. a slice is a class * list of selectors. aBlock will be applied to all the extensions slices of the receiver. aBlok first argument is the class and the second argument a list of method selectors"
	
	classDefinedSelectors keysAndValuesDo: aBlock! !

!RPackage methodsFor: 'slices' stamp: 'DamienPollet 10/1/2010 11:48'!
classExtensionSlicesDo: aBlock
	"This method iterates over the class extensions and their associated selectors. A slice is a class * list of selectors. aBlock will be applied to all the extensions slices of the receiver. aBlock first argument is the class and the second argument a list of method selectors"
	
	classExtensionSelectors keysAndValuesDo: aBlock! !

!RPackage methodsFor: 'slices' stamp: 'StephaneDucasse 3/11/2010 18:40'!
classNamesAndExtensionSelectorsDo: aBlock
	"Iterate over the extended methods grouped by classes and apply the argument.
	The block will be passed a class name and each individual selectors.
	Only classes with methods are paired with their methods"
	
	classExtensionSelectors keysAndValuesDo: [ :classSymbol :methods |
			methods do: [ :sel |  aBlock value: classSymbol	value: sel]].
	
	metaclassExtensionSelectors keysAndValuesDo: [ :classSymbol :methods |
			methods do: [ :sel |  aBlock 
									value: (classSymbol, '_class') asSymbol
									value: sel]].	! !

!RPackage methodsFor: 'all accessing' stamp: 'sd 4/25/2010 16:03'!
classes
	"Return all the classes"
	
	^ self definedClasses, self extendedClasses 
	
	! !

!RPackage methodsFor: 'all accessing' stamp: 'StephaneDucasse 3/11/2010 15:17'!
definedClassNames 
	"Return the class names having methods defined in the receiver."
	
	^ classes! !

!RPackage methodsFor: 'all accessing' stamp: 'StephaneDucasse 5/15/2010 19:47'!
definedClasses 
	
	^ classes collect: [:each | Smalltalk globals at: each]
	! !

!RPackage methodsFor: 'iteration' stamp: 'StephaneDucasse 2/21/2010 17:25'!
definedClassesDo: aBlock

	classDefinedSelectors keysDo: aBlock.
	metaclassDefinedSelectors keysDo: aBlock.! !

!RPackage methodsFor: 'private' stamp: 'sd 4/25/2010 15:29'!
definedMethodsBecomeExtendedForClass: aClassName
	"the package may contain defined methods and their class is removed to the receiver. The status of such methods should now be extended"
	"Precondition: aClassName is in the defined list"
	
	| ext |
	(classDefinedSelectors keys includes: aClassName) ifFalse: [^ self].
	ext := classDefinedSelectors at: aClassName.
	classDefinedSelectors removeKey: aClassName.
	(classExtensionSelectors keys includes: aClassName) ifTrue: [self error: aClassName , ' should not be defined in extension'.]. "we should remove this check later"
	classExtensionSelectors at: aClassName put: ext.! !

!RPackage methodsFor: 'private' stamp: 'cyrille.delaunay 11/24/2010 14:14'!
definedMethodsBecomeExtendedForMetaclass: aClassName
	"the package may contain defined methods and their class is removed to the receiver. The status of such methods should now be extended"
	"Precondition: aClassName is in the defined list"
	
	| ext |
	(metaclassDefinedSelectors keys includes: aClassName) ifFalse: [^ self].
	ext := metaclassDefinedSelectors at: aClassName.
	metaclassDefinedSelectors removeKey: aClassName.
	(metaclassExtensionSelectors keys includes: aClassName) ifTrue: [self error: aClassName , ' should not be defined in extension'.]. "we should remove this check later"
	metaclassDefinedSelectors at: aClassName put: ext.! !

!RPackage methodsFor: 'accessing' stamp: 'tg 3/13/2010 14:17'!
definedMethodsForClass: aClass

	^ (self definedSelectorsForClass: aClass) collect: [ :each | aClass >> each ]! !

!RPackage methodsFor: 'private' stamp: 'sd 4/25/2010 15:41'!
definedMethodsShouldBecomeExtensionWhenRemovingClass: aClassName
	"the package may contain defined methods and their class is removed from the receiver. The status of such methods should now be extensions"
	
	self definedMethodsBecomeExtendedForClass: aClassName.
	self definedMethodsBecomeExtendedForMetaclass: aClassName.! !

!RPackage methodsFor: 'accessing' stamp: 'tg 3/13/2010 13:31'!
definedSelectorsForClass: aClass

	^ aClass isMeta
		ifTrue: [ metaclassDefinedSelectors at: aClass theNonMetaClass name ifAbsent: [ #() ] ]
		ifFalse: [ classDefinedSelectors at: aClass name ifAbsent: [ #() ] ]! !

!RPackage methodsFor: 'testing' stamp: 'StephaneDucasse 4/20/2010 22:22'!
definesClass: aClass
	"Returns true if the class, aClass, is one of the locally defined classes"
	"should be probably removed since this is the same as includesClass: or the inverse"
	^ self includesClass: aClass! !

!RPackage methodsFor: 'testing' stamp: 'sd 4/25/2010 16:06'!
definesOrExtendsClass: aClass
	"Returns true whether the class, aClass, is one of the locally defined classes of the receiver or 
	if the receiver extends such class (that is defined in another package)"
	
	| canonizedName |
	canonizedName := aClass theNonMetaClass name.
	^ (classes includes: canonizedName) 
		or: [self extendedClassNames includes:  canonizedName]
! !

!RPackage methodsFor: 'accessing' stamp: 'sd 4/25/2010 16:06'!
extendedClassNames 
	"Return the name of the classes which are extended by the receiver package. if a metaclass is extended, just get its sole instance class name."
	
	^ classExtensionSelectors keys union: metaclassExtensionSelectors keys! !

!RPackage methodsFor: 'accessing' stamp: 'sd 4/25/2010 16:03'!
extendedClasses 
	"Return classes and metaclasses that are extended in the receiver"
	
	^ (metaclassExtensionSelectors keys collect: [:each | (Smalltalk at: each) class])
			union:  (classExtensionSelectors keys collect: [:each | Smalltalk at: each])! !

!RPackage methodsFor: 'private' stamp: 'sd 4/25/2010 15:29'!
extendedMethodsBecomeDefinedForClass: aClassName
	"the package may contain extension methods and their class is added to the receiver. The status of such methods should now be defined"
	"Precondition: aClassName is in the defined list"
	
	| ext |
	(classExtensionSelectors keys includes: aClassName) ifFalse: [^ self].
	ext := classExtensionSelectors at: aClassName.
	classExtensionSelectors removeKey: aClassName.
	(classDefinedSelectors keys includes: aClassName) ifTrue: [self error: aClassName , ' should not be defined in extension'.]. "we should remove this check later"
	classDefinedSelectors at: aClassName put: ext.! !

!RPackage methodsFor: 'private' stamp: 'StephaneDucasse 5/15/2010 22:22'!
extendedMethodsBecomeDefinedForMetaclass: aClassName
	"the package may contain extension methods and their class is added to the receiver. The status of such methods should now be defined"
	"Precondition: aClassName is in the defined list"
	
	| ext |
	(metaclassExtensionSelectors keys includes: aClassName) ifFalse: [^ self].
	ext := metaclassExtensionSelectors at: aClassName.
	metaclassExtensionSelectors removeKey: aClassName.
	(metaclassDefinedSelectors keys includes: aClassName) 
			ifTrue: [self error: aClassName , ' should not be defined in extension'.]. "we should remove this check later"
	metaclassDefinedSelectors at: aClassName put: ext.! !

!RPackage methodsFor: 'private' stamp: 'sd 4/25/2010 15:29'!
extendedMethodsShouldBecomeDefinedWhenAddingClassName: aClassName
	"the package may contain extension methods and their class is added to the receiver. The status of such methods should now be defined"
	"Precondition: aClassName is in the defined list"
	
	self extendedMethodsBecomeDefinedForClass: aClassName.
	self extendedMethodsBecomeDefinedForMetaclass: aClassName.
	! !

!RPackage methodsFor: 'testing' stamp: 'sd 4/25/2010 16:05'!
extendsClass: aClass
	"Returns true if the receiver extends aClass (that is defined in another package)"
	
	| canonizedName |
	canonizedName := aClass theNonMetaClass name.
	^ self extendedClassNames includes:  canonizedName! !

!RPackage methodsFor: 'accessing' stamp: 'cyrille.delaunay 10/27/2010 13:04'!
extensionMethods
	"Extension methods are methods defined on classes that are not defined in the receiver"
	
	| mthds |
	mthds := OrderedCollection new.
	classExtensionSelectors keysAndValuesDo: [ :classSymbol :methods |
			methods do: [ :sel | mthds add: ((Smalltalk at: classSymbol) >>	sel)]].
	metaclassExtensionSelectors
	 keysAndValuesDo: [ :classSymbol :methods |
			methods do: [ :sel | mthds add: ((Smalltalk at: classSymbol) class >>sel)]].	
	^ mthds				! !

!RPackage methodsFor: 'accessing' stamp: 'tg 3/13/2010 14:18'!
extensionMethodsForClass: aClass

	^ (self extensionSelectorsForClass: aClass) collect: [ :each | aClass >> each ]! !

!RPackage methodsFor: 'accessing' stamp: 'DamienPollet 9/28/2010 17:09'!
extensionSelectors
	"Extension methods are methods defined on classes that are not defined in the receiver"
	
	| classSels |
	classSels := Set new.
	classExtensionSelectors values do: [:each | classSels addAll: each].
	metaclassExtensionSelectors values do: [:each | classSels addAll: each].
	^ classSels.
	! !

!RPackage methodsFor: 'accessing' stamp: 'StephaneDucasse 4/21/2010 14:05'!
extensionSelectorsForClass: aClass

	^ aClass isMeta
		ifTrue: [ metaclassExtensionSelectors at: aClass theNonMetaClass name ifAbsent: [ #() ] ]
		ifFalse: [ classExtensionSelectors at: aClass name ifAbsent: [ #() ] ]! !

!RPackage methodsFor: 'testing' stamp: 'StephaneDucasse 2/21/2010 17:56'!
includesClass: aClass	
	"Returns true if the receiver includes aClass in the classes that are defined within it: only class definition are considered - not class extensions"
	
	^ self includesClassNamed: aClass theNonMetaClass name! !

!RPackage methodsFor: 'testing' stamp: 'StephaneDucasse 4/20/2010 22:53'!
includesClassNamed: aSymbol
	"Returns true if the receiver includes class named aSymbol in the classes that are defined within it: only class definition are considered - not class extensions"
	
	| aRealSymbol |
	aRealSymbol := aSymbol asSymbol.
	^ classes includes: aSymbol
	! !

!RPackage methodsFor: 'testing' stamp: 'tg 3/13/2010 15:13'!
includesDefinedSelector: aSelector ofClass: aClass
	"Return true if the receiver includes the method of selector aSelector. Only checks methods defined in this package"
	
	| sels |
	sels := self definedSelectorsForClass: aClass.
	^ sels includes: aSelector.! !

!RPackage methodsFor: 'testing' stamp: 'StephaneDucasse 2/21/2010 18:53'!
includesDefinedSelector: aSelector ofClassName: aClassName
	"Return true if the receiver includes the method of selector aSelector. Only checks methods defined (not extended by other packages or package extensions)"
	
	| sels |
	sels := classDefinedSelectors at: aClassName ifAbsent: [^ false].
	^ sels includes: aSelector.! !

!RPackage methodsFor: 'testing' stamp: 'StephaneDucasse 2/21/2010 18:53'!
includesDefinedSelector: aSelector ofMetaclassName: aClassName
	"Return true if the receiver includes the method of selector aSelector. Only checks methods defined (not extended by other packages or package extensions)"
	
	| sels |
	sels := metaclassDefinedSelectors at: aClassName ifAbsent: [^ false].
	^ sels includes: aSelector.! !

!RPackage methodsFor: 'testing' stamp: 'tg 3/13/2010 13:47'!
includesExtensionSelector: aSelector ofClass: aClass
	"Return true if the receiver includes the method of selector aSelector. Only checks methods extending other packages"
	
	| sels |
	sels := self extensionSelectorsForClass: aClass.
	^ sels includes: aSelector.! !

!RPackage methodsFor: 'testing' stamp: 'StephaneDucasse 2/21/2010 18:54'!
includesExtensionSelector: aSelector ofClassName: aClassName
	"Return true if the receiver includes the method of selector aSelector. Only checks methods extending other packages"
	
	| sels |
	sels := classExtensionSelectors at: aClassName ifAbsent: [^ false].
	^ sels includes: aSelector.! !

!RPackage methodsFor: 'testing' stamp: 'StephaneDucasse 2/21/2010 18:54'!
includesExtensionSelector: aSelector ofMetaclassName: aClassName
	"Return true if the receiver includes the method of selector aSelector. Only checks methods extending other packages"
	
	| sels |
	sels := metaclassExtensionSelectors at: aClassName ifAbsent: [^ false].
	^ sels includes: aSelector.! !

!RPackage methodsFor: 'testing' stamp: 'tg 3/13/2010 16:42'!
includesSelector: aSelector ofClass: aClass
	"Return true if the receiver includes the method of selector aSelector. Checks methods defined locally as well as extending other packages"
	
	^ aClass isMeta
		ifTrue: [ self includesSelector: aSelector ofMetaclassName: aClass theNonMetaClass name  ]
		ifFalse: [ self includesSelector: aSelector ofClassName: aClass name  ]
	
! !

!RPackage methodsFor: 'testing' stamp: 'StephaneDucasse 2/21/2010 20:50'!
includesSelector: aSelector ofClassName: aClassName
	"Return true if the receiver includes the method of selector aSelector. Checks methods defined locally as well as extending other packages"
	
	^ (self includesDefinedSelector: aSelector ofClassName: aClassName)
		or: [(self includesExtensionSelector: aSelector ofClassName: aClassName)]! !

!RPackage methodsFor: 'testing' stamp: 'StephaneDucasse 3/11/2010 17:38'!
includesSelector: aSelector ofMetaclassName: aClassName
	"Return true if the receiver includes the method of selector aSelector. Checks methods defined locally as well as extending other packages"
	
	^ (self includesDefinedSelector: aSelector ofMetaclassName: aClassName)
		or: [(self includesExtensionSelector: aSelector ofMetaclassName: aClassName)]! !

!RPackage methodsFor: 'initialize-release' stamp: 'StephaneDucasse 3/11/2010 15:09'!
initialize
	super initialize.
	classes := Set new.
	metaclassExtensionSelectors := Dictionary new.
	classExtensionSelectors := Dictionary new.
	classDefinedSelectors := Dictionary new.
	metaclassDefinedSelectors := Dictionary new.! !

!RPackage methodsFor: 'slices' stamp: 'StephaneDucasse 3/11/2010 18:47'!
metaclassDefinedSlicesDo: aBlock
	"This method iterates over the defined class and their associated selectors. a slice is a class * list of selectors. aBlock will be applied to all the extensions slices of the receiver. aBlok first argument is the class and the second argument a list of method selectors"
	
	metaclassDefinedSelectors keysAndValuesDo: aBlock! !

!RPackage methodsFor: 'slices' stamp: 'StephaneDucasse 4/20/2010 21:42'!
metaclassExtensionSlicesDo: aBlock
	"This method iterates over the metaclass extensions and their associated selectors. A slice is a class * list of selectors. aBlock will be applied to all the extensions slices of the receiver. aBlok first argument is the class and second argument a list of selectors"
	
	metaclassExtensionSelectors keysAndValuesDo: aBlock
	
! !

!RPackage methodsFor: 'all accessing' stamp: 'sd 4/21/2010 16:36'!
methods

	| mths |
	mths := OrderedCollection new.
	metaclassExtensionSelectors keysAndValuesDo: [:key :val | val do: [:sel | mths add: ((Smalltalk at: key) class >> sel)]].
	classExtensionSelectors keysAndValuesDo: [:key :val |val do: [:sel | mths add: ((Smalltalk at: key) >> sel)]].
	metaclassDefinedSelectors keysAndValuesDo: [:key :val | val do: [:sel | mths add: ((Smalltalk at: key) class >> sel)]].
	classDefinedSelectors keysAndValuesDo: [:key :val | val do: [:sel | mths add: ((Smalltalk at: key) >> sel)]].
	^ mths! !

!RPackage methodsFor: 'accessing' stamp: 'DamienPollet 9/28/2010 17:10'!
methodsForClass: aClass
	"Returns all compiled methods for a class in the receiver without distinction of extensions or not"
	
	^ (self includesClass: aClass)
		ifFalse: [self extensionMethodsForClass: aClass]
		ifTrue: [self definedMethodsForClass: aClass]
	! !

!RPackage methodsFor: 'accessing' stamp: 'sd 4/25/2010 15:52'!
name

	^ name! !

!RPackage methodsFor: 'accessing' stamp: 'sd 4/25/2010 15:52'!
name: aSymbol

	name := aSymbol asSymbol! !

!RPackage methodsFor: 'accessing' stamp: 'DamienPollet 10/1/2010 11:52'!
organizer
	^ self class organizer! !

!RPackage methodsFor: 'printing' stamp: 'sd 4/25/2010 15:51'!
printOn: aStream
	super printOn: aStream.
	aStream nextPut: Character space.
	aStream nextPutAll: self name.! !

!RPackage methodsFor: 'private-register' stamp: 'cyrille.delaunay 11/17/2010 13:45'!
registerClass: aClass
	self organizer
		registerPackage: self forClass: aClass
	! !

!RPackage methodsFor: 'private-register' stamp: 'cyrille.delaunay 11/25/2010 11:25'!
registerClassName: aClassNameSymbol

	self organizer
		registerPackage: self forClassName: aClassNameSymbol
	! !

!RPackage methodsFor: 'class addition' stamp: 'sd 4/25/2010 15:47'!
removeClassDefinition: aClass
	"Remove the class definition from the package but not its methods"
	
	| removed aClassName|
	aClassName := aClass theNonMetaClass name.
	removed := classes remove: aClassName ifAbsent: [nil].
	removed ifNotNil: [
			self definedMethodsShouldBecomeExtensionWhenRemovingClass: aClassName.
			self unregisterClass: aClass.] 
	! !

!RPackage methodsFor: 'class addition' stamp: 'sd 4/25/2010 15:46'!
removeClassDefinitionName: aClassName
	"remove the class definition from the package but not its method"
	
	| removed aClassNameSymbol |
	('* class' match: aClassName) ifTrue: [^ self error: 'no metaclass name'].
	aClassNameSymbol := aClassName asSymbol.
	removed := classes remove: aClassNameSymbol ifAbsent: [nil].
	removed ifNotNil: [
				self definedMethodsShouldBecomeExtensionWhenRemovingClass: aClassName.
				self unregisterClassName: aClassNameSymbol.] 
	! !

!RPackage methodsFor: 'compiled method handling' stamp: 'DamienPollet 10/1/2010 11:54'!
removeMethod: aCompiledMethod 
	"Remove the method to the receiver as a defined method."

	| mr mthClass |
	mthClass := aCompiledMethod methodClass. 
	(self includesClass: mthClass)
		ifFalse: [mthClass isMeta
					ifTrue: [(metaclassExtensionSelectors 
								at: mthClass theNonMetaClass name  
								ifAbsentPut: [Set new]) remove: aCompiledMethod selector.]
					ifFalse: [(classExtensionSelectors 
								at: mthClass name 
								ifAbsentPut: [Set new]) remove: aCompiledMethod selector].
				"we remove a method extension so the receiver is an extending package of the class"
				self organizer unregisterExtendingPackage: self forClass: mthClass]
		ifTrue: [mthClass isMeta
					ifTrue: [(metaclassDefinedSelectors 
								at: mthClass theNonMetaClass name  
								ifAbsentPut: [Set new]) remove: aCompiledMethod selector]
					ifFalse: [(classDefinedSelectors 
								at: mthClass name 
								ifAbsentPut: [Set new]) remove: aCompiledMethod selector]].	
	^ aCompiledMethod.! !

!RPackage methodsFor: 'selector handling' stamp: 'StephaneDucasse 4/20/2010 22:57'!
removeSelector: aSelector ofClassName: aClassName
	"Remove the method in the package but does not propagate to the class itself.  Note that this method does not remove the method from the class, it just records in the package that the method is not in the package anymore."
	
	( self includesClassNamed: aClassName)
		ifFalse: [(classExtensionSelectors  at: aClassName ifAbsent: [ ^ Set new]) remove: aSelector.]
		ifTrue: [(classDefinedSelectors  at: aClassName ifAbsent: [ ^ Set new]) remove: aSelector.].
	
	! !

!RPackage methodsFor: 'selector handling' stamp: 'StephaneDucasse 4/20/2010 22:57'!
removeSelector: aSelector ofMetaclassName: aClassName
	"Remove the method in the package. Note that this method does not remove the method from the class, it just records in the package that the method is not in the package anymore. aClassName is the sole instance class name and not its metaclass one: i.e. adding Point class>>new is done as removeSelector: #new ofMetaclassName: #Point"
	
	( self includesClassNamed: aClassName)
		ifFalse: [(metaclassExtensionSelectors  at: aClassName ifAbsent: [ ^ Set new]) remove: aSelector.]
		ifTrue: [(metaclassDefinedSelectors  at: aClassName ifAbsent: [ ^ Set new]) remove: aSelector.].
	
	! !

!RPackage methodsFor: 'all accessing' stamp: 'sd 4/21/2010 16:12'!
selectors

	| sel |
	sel := Set new.
	metaclassExtensionSelectors valuesDo: [:each | sel addAll: each].
	classExtensionSelectors valuesDo: [:each | sel addAll: each].
	metaclassDefinedSelectors valuesDo: [:each | sel addAll: each].
	classDefinedSelectors valuesDo: [:each | sel addAll: each].
	^ sel
	! !

!RPackage methodsFor: 'accessing' stamp: 'DamienPollet 9/28/2010 17:11'!
selectorsForClass: aClass
	"Returns all selectors for a class in the receiver without distinction of extensions or not"
	
	^ (self includesClass: aClass)
		ifFalse: [self extensionSelectorsForClass: aClass]
		ifTrue: [self definedSelectorsForClass: aClass]
	! !

!RPackage methodsFor: 'private-register' stamp: 'cyrille.delaunay 11/17/2010 13:46'!
unregisterClass: aClass
	self organizer
		unregisterPackage: self forClass: aClass.! !

!RPackage methodsFor: 'private-register' stamp: 'cyrille.delaunay 11/17/2010 13:46'!
unregisterClassName: aClassNameSymbol
	self organizer unregisterPackage: self forClassName: aClassNameSymbol! !

!RPackage methodsFor: 'updating' stamp: 'cyrille.delaunay 11/16/2010 10:29'!
updateClassDefinedNamed: oldString withNewName: newString

	"=> update the 'classDefinedSelectors' dictionary (replace the old key by the new one)
       => update the 'metaclassDefinedSelectors' dictionary (replace the old key by the new one)"
	classes remove: oldString.
	classes add: newString.
	
	(classDefinedSelectors at: oldString ifAbsent: [nil]) ifNotNil: [
		classDefinedSelectors at: newString put: (classDefinedSelectors at: oldString).
		classDefinedSelectors removeKey: oldString
		].
	
	(metaclassDefinedSelectors at: oldString ifAbsent: [nil]) ifNotNil: [
		metaclassDefinedSelectors at: newString put: (metaclassDefinedSelectors at: oldString).
		metaclassDefinedSelectors removeKey: oldString
		].! !

!RPackage methodsFor: 'updating' stamp: 'cyrille.delaunay 11/16/2010 14:40'!
updateDefinedSelector: oldSelector inClass: aClass withNewSelector: newSelector

	aClass isMeta
		ifTrue: [
			(metaclassDefinedSelectors at: aClass instanceSide name asSymbol) remove: oldSelector. 
			(metaclassDefinedSelectors at: aClass instanceSide name asSymbol) add: newSelector. 
			]
		ifFalse: [
			(classDefinedSelectors  at: aClass name asSymbol) remove: oldSelector. 
			(classDefinedSelectors  at: aClass name asSymbol) add: newSelector. 
			]! !

!RPackage methodsFor: 'updating' stamp: 'cyrille.delaunay 11/24/2010 14:04'!
updateExtensionForClassNamed: oldString withNewName: newString

	"=> update the 'classExtensionsSelectors' dictionary (replace the old key by the new one)
        => update the 'metaclassclassExtensionsSelectors' dictionary (replace the old key by the new one)"
	"classes remove: oldString.
	classes add: newString."
	
	(metaclassExtensionSelectors at: oldString ifAbsent: [nil]) ifNotNil: [
		metaclassExtensionSelectors at: newString put: (metaclassExtensionSelectors at: oldString).
		metaclassExtensionSelectors removeKey: oldString
		].
	
	(classExtensionSelectors at: oldString ifAbsent: [nil]) ifNotNil: [
		classExtensionSelectors at: newString put: (classExtensionSelectors at: oldString).
		classExtensionSelectors removeKey: oldString
		].! !

!RPackage methodsFor: 'updating' stamp: 'cyrille.delaunay 11/16/2010 14:41'!
updateExtensionSelector: oldSelector inClass: aClass withNewSelector: newSelector

	aClass isMeta
		ifTrue: [
			(metaclassExtensionSelectors at: aClass instanceSide name asSymbol) remove: oldSelector. 
			(metaclassExtensionSelectors at: aClass instanceSide name asSymbol) add: newSelector. 
			]
		ifFalse: [
			(classExtensionSelectors  at: aClass name asSymbol) remove: oldSelector. 
			(classExtensionSelectors  at: aClass name asSymbol) add: newSelector. 
			]! !

!RPackage methodsFor: 'updating' stamp: 'cyrille.delaunay 11/16/2010 15:08'!
updateSelector: oldSelector inClass: aClass withNewSelector: newSelector

	"here we check if the old seletor is defined in the package or if it is an extension. Then we dispatch the work to the corresponding method.
	For one specific class, a selector is either a defined selector or an extension selector, there is no way to have a couple of class->selector present in both dictionary"
	
	(self includesDefinedSelector: oldSelector ofClass: aClass) 
		ifTrue: [^ self updateDefinedSelector: oldSelector inClass: aClass  withNewSelector: newSelector].
	(self includesExtensionSelector: oldSelector ofClass: aClass)
		ifTrue: [^ self updateExtensionSelector: oldSelector inClass: aClass  withNewSelector: newSelector].
		
		! !

Object subclass: #RPackageBenchmark
	instanceVariableNames: ''
	classVariableNames: 'OrganizerToBenchmark'
	poolDictionaries: ''
	category: 'RPackage-Core'!

!RPackageBenchmark class methodsFor: 'as yet unclassified' stamp: 'StephaneDucasse 7/16/2010 15:58'!
organizerToBenchmark
	"self organizerToBenchmark"
	OrganizerToBenchmark 
		ifNil: [ OrganizerToBenchmark := RPackageOrganizer new initializeFromSystem].
	^ OrganizerToBenchmark! !

!RPackageBenchmark class methodsFor: 'as yet unclassified' stamp: 'StephaneDucasse 7/16/2010 15:53'!
resetOrganizerToBenchmark

	OrganizerToBenchmark := nil! !

!RPackageBenchmark class methodsFor: 'as yet unclassified' stamp: 'StephaneDucasse 7/29/2010 23:00'!
runForLog
	"To run the benchmark you may have to reset the organizer
	self resetOrganizerToBenchmark"	
	" self runForLog  "
	
	
	"17 July 2010
		class package access: 10
		extending package access: 13
		selector for class: 39
		extending package access +: 19
		method package access: 1173
	"
	

	
	| classes |
	RPackage 
		withOrganizer: self organizerToBenchmark
		do: [ :anOrganizer | 
			classes := Object withAllSubclasses.
			classes remove: UndefinedObject.
			classes remove: ClassTrait.
			classes remove: ClassTrait class.
			classes := classes reject: [:each | #('BogusInfo*' 'GRPackage*' 'MC*') 
												anySatisfy: [:pat | pat match: each name]].
			classes := classes reject: [:each | each package isNil].
			
			^ String streamContents: [:st |
					
					st nextPutAll: 'class package access: ',	(Time millisecondsToRun: [classes do: [:each | each package]]) printString ;cr.
					st nextPutAll: 'extending package access: ',	(Time millisecondsToRun: [classes do: [:each | each extendingPackages]]) printString ;cr.
		     			st nextPutAll: 'selector for class: ',	(Time millisecondsToRun: [classes do: [:each | each package selectorsForClass: each]]) printString ;cr.
					st nextPutAll: 'extending package access +: ',	(Time millisecondsToRun: 
																	[classes do: [:each | each extendingPackages 
																							do: [ :package | 
																								package selectorsForClass: each]]]) printString ;cr.
					st nextPutAll: 'method package access: ',	(Time millisecondsToRun: 
																	[classes do: [:each | each methods do: [:method | 
																				method packageFromOrganizer: anOrganizer ]]]) printString ;cr.
			
			
			
			]].
					
	
	! !

Object subclass: #RPackageOrganizer
	instanceVariableNames: 'classPackageMapping packages classExtendingPackagesMapping'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'RPackage-Core'!

RPackageOrganizer class
	instanceVariableNames: 'default'!

!RPackageOrganizer commentStamp: '<historical>' prior: 0!
A RPackageOrganizer is responsible for providing all the package currently defined in the system.
In addition it provides a back pointer from the class to its package.

The classPackageMapping and the classExtendingPackageMapping should be moved in the future to the classes themselves. 


For tests or actions that could destroy the package organizer,  do not access directly the singleton of RPackageOrganizer.
Use instead 
		RPackage withOrganizer: aNewOrganizer do: ablock
			or via RPackage organizer
		
	
RPackageOrganizer fillUp will fill up the system from the current PackageOrganizer
	"self fillUp"
	
-----------------------------------------------------------------------------------------------------------------------------------------------	
	
	
A rpackageOrganizer update itself when some changes are made in the system. It does that by registering to a systemAnnoucer, specifying an action when an annoucement is triggered.
Here is what I (Cyrille Delaunay) propose to do for each annocuement triggered:

SystemCategoryAddedAnnouncement 	
	=> I would just register a new RPackage (if it does not already exist) in the RPackageOrganizer
	
SystemCategoryRemovedAnnouncement 
     => I would just unregister the RPackage concerned from the organizer

SystemCategoryRenamedAnnouncement
     => I would update the RPackage concerned, by changing its name
     => I would update the 'packages' dictionary of the organizer, putting the new name as key

SystemClassAddedAnnouncement 
    => Import the class in the RPackage concerned (RPackage >> importClass:)
    => Register the class in the 'classPackageMapping' dictionary of the organizer (RPackageOrganizer >> registerPackage:forClass)
    (=> maybe we should pay attention if both the class and the metaclass launch this kind of event ?)

SystemClassRecategorizedAnnouncement
    => I would update the old RPackage concerned:
            => unregister the class
            => unregister all defined methods of the class
    => I would update the new RPackage:
            => Import the class in the RPackage (importClass:)
    => I would update the organizer:
            => update the 'classPackageDictionary' to point on the new RPackage

 
SystemClassRemovedAnnouncement
    => I would update the RPackake concerned
             => unregister the class
             => unregister all defined methods of the class
    => I would update the organizer:
             => update the 'classPackageDictionary' to remove the class

SystemClassRenamedAnnouncement
    => I would update the RPackage in which the class is defined:
             => update the 'classDefinedSelectors' dictionary (replace the old key by the new one)
             => update the 'metaclassDefinedSelectors' dictionary (replace the old key by the new one)
    => I would update all RPackages extending this class
             => update the 'classExtensionsSelectors' dictionary (replace the old key by the new one)
             => update the 'metaclassclassExtensionsSelectors' dictionary (replace the old key by the new one)
    => I would update the organizer
             => update the 'classPackageDictionary' to replace the key with the new class name
             => update the 'classExtendingPackagesMapping' to replace the key with the new class name
                          
SystemClassReorganizedAnnouncement 
    (=> I guess we should check if extensions have not been added or removed ? 
      (to retrieve this information, the only thing I found is ClassDescription >> organization, and then check each category begining with '*' and compare with the organizer. seems to be painful, no?))
	=> when an extension is removed, all methods inside are removed. Therefore, the MethodRemovedAnnounecement will do the job. Not sur this one still usefull

SystemProtocolAddedAnnouncement
    => I don't see anything to do for this annoucement

SystemProtocolRemovedAnnoucement
    => If the category is an extension from a package, I would move all the methods concerned, from the extending RPackage to the class RPackage

SystemMethodAddedAnnouncement
       => I would check the category in which the method has been defined
               => if it correspond to an extending package -> add the method to the extending RPackage
               => if not, add the method to the class parentPackage

SystemMethodModifiedAnnouncement
       this annoucement can correspond to several kind of modifications:
	       *  a method has been renamed
                       => I would update the rPackage in which the method is defined to replace the old selector by the new one
		* a method has been move to another category 
			-maybe from a classic category to an extending package
                             => we should move the method from the  method class parentPackage to extendingPackage package
			-maybe from an extending package to another extending package
                             => we should move the method from the  extendingPackage package to the other extendingPackage package
			-maybe from an extending package to a classic category
                             =>  we should move the method from the  extendingPackage to the method class parentPackage
		        -maybe from a classic category to another classic category
                             => we have nothing to do
			

SystemMethodRecategorizedAnnouncement
          same thing than above

SystemMethodRemovedAnnouncement
       => I would simply remove the method from the RPackage in which it is register!

RPackageOrganizer class
	instanceVariableNames: 'default'!

!RPackageOrganizer class methodsFor: 'utils' stamp: 'cyrille.delaunay 11/3/2010 18:26'!
computeDiffWithMC
	"self computerDiffWithMC"
	
	^ self rpackageNames difference: self packageInfoNames ! !

!RPackageOrganizer class methodsFor: 'singleton' stamp: 'TudorGirba 11/25/2010 23:51'!
default
	
	default isNil ifTrue: [
		default := self new.
		default ].
	^ default! !

!RPackageOrganizer class methodsFor: 'singleton' stamp: 'TudorGirba 11/25/2010 23:56'!
fillUp
	"self fillUp"
	self default initialize.
	self default initializeFromSystem..! !

!RPackageOrganizer class methodsFor: 'instance creation' stamp: 'cyrille.delaunay 11/18/2010 14:16'!
new

	^ super new.! !

!RPackageOrganizer class methodsFor: 'utils' stamp: 'cyrille.delaunay 11/3/2010 18:25'!
packageInfoNames
	"self packageInfoNames"
	
	^ MCWorkingCopy allManagers collect: [:each | each package name]
	
	! !

!RPackageOrganizer class methodsFor: 'utils' stamp: 'cyrille.delaunay 11/3/2010 18:24'!
rpackageNames
	"self computeDiffWithMC"
	
	^ default packages collect: [:each | each name]
	
	! !

!RPackageOrganizer methodsFor: 'mc synchronization' stamp: 'StephaneDucasse 7/29/2010 12:00'!
category: categoryName matches: prefix
	"self default category: 'Compression-Archives' matches: 'Compression'"
	"self default category: 'Compression' matches: 'Compression-Archives'"
	"self default category: 'Compression' matches: 'Compression'"
	"self default category: 'Compression-Arh' matches: 'Compression'"
	| prefixSize catSize |
	categoryName ifNil: [ ^false ].
	catSize := categoryName size.
	prefixSize := prefix size.
	catSize < prefixSize ifTrue: [ ^false ].
	(categoryName findString: prefix startingAt: 1 caseSensitive: false) = 1
		ifFalse: [ ^false ].
	^(categoryName at: prefix size + 1 ifAbsent: [ ^true ]) = $-! !

!RPackageOrganizer methodsFor: 'initialize' stamp: 'TudorGirba 11/26/2010 00:27'!
defineUnpackagedClassesPackage
	^ self registerPackageNamed: #UnpackagedClasses! !

!RPackageOrganizer methodsFor: 'package access from class' stamp: 'sd 5/21/2010 14:56'!
extendingPackagesOf: aClass
	"Returns the packages extending the class aClass"
	^ classExtendingPackagesMapping at: aClass theNonMetaClass name ifAbsent: [#()]! !

!RPackageOrganizer methodsFor: 'package access from class' stamp: 'cyrille.delaunay 11/24/2010 14:24'!
extendingPackagesOfClassNamed: aName
	"Returns the packages extending the class named a Symbol"
	^ classExtendingPackagesMapping at: aName asSymbol ifAbsent: [#()]! !

!RPackageOrganizer methodsFor: 'testing' stamp: 'sd 5/21/2010 15:21'!
includesClass: aClass

	^ self includesClassNamed: aClass name! !

!RPackageOrganizer methodsFor: 'testing' stamp: 'sd 5/21/2010 15:18'!
includesClassNamed: aSymbol

	^ classPackageMapping keys includes: aSymbol ! !

!RPackageOrganizer methodsFor: 'testing' stamp: 'sd 5/21/2010 15:17'!
includesPackage: aPackage

	^ self includesPackageNamed: aPackage name! !

!RPackageOrganizer methodsFor: 'testing' stamp: 'sd 5/21/2010 15:16'!
includesPackageNamed: aSymbol

	^ packages keys includes:  aSymbol ! !

!RPackageOrganizer methodsFor: 'initialize' stamp: 'TudorGirba 11/26/2010 00:20'!
initialize
	super initialize.
	packages := IdentityDictionary new.
	classPackageMapping := IdentityDictionary new.
	classExtendingPackagesMapping := IdentityDictionary new.
	self register.
	self defineUnpackagedClassesPackage! !

!RPackageOrganizer methodsFor: 'initialize' stamp: 'cyrille.delaunay 11/3/2010 18:16'!
initializeFromSystem
	""
	| package |
	MCWorkingCopy allManagers 
		do: [ :workingCopy |
			| info |
			info := workingCopy packageInfo.
			"self haltIf: [info packageName = 'Moose']."
			package := self registerPackageNamed: info packageName.
			"My problem here: If a package appear here after one of its subcategories, all classes and methods would be re-affected to this package instead of the subcatagory. That's why I focus only on classes and methods that are not defined in subcategories, to avoid to override information set before"
			info classesNotDefinedInSubCategories  
				do: [ :each | package addClassDefinition: each ].
			info coreMethodsNotDefinedInSubCategories do: [ :each |
				each isValid
					ifTrue: [ package addMethod: each compiledMethod ] ].
			info extensionMethods  do: [ :each |
				each isValid
					ifTrue: [ package addMethod: each compiledMethod ] ] ]
		displayingProgress: 'Importing from MC managers'
	"strange PackageOrganizer default packages size is not equal to MCWorkingCopy allManagers size
	MCWorkingCopy allManagers size = PackageOrganizer default packages size
	"
	
	"PackageOrganizer default packages do: [ :info |
			package := RPackage named: info name.
			info classes do: [ :each | package addClassDefinition: each ].
			info coreMethods do: [ :each |
				each isValid
					ifTrue: [ package addMethod: each compiledMethod ] ].
			info extensionMethods do: [ :each |
				each isValid
					ifTrue: [ package addMethod: each compiledMethod ] ] ]
		displayingProgress: 'Importing'"! !

!RPackageOrganizer methodsFor: 'mc synchronization' stamp: 'StephaneDucasse 7/29/2010 13:14'!
isDefinedAsPackageOrSubPackageInMC: aSymbol
	"a category has been added. 
	#'AST-Core'
	Now the question is to know whether AST-Core already exists as a package.
	"
	"self default isDefinedAsPackageOrSubPackageInMC: #Compression"
	"self default isDefinedAsPackageOrSubPackageInMC: #'Compression-Archives'"
	"self default isDefinedAsPackageOrSubPackageInMC: #'Collections'"
	"self default isDefinedAsPackageOrSubPackageInMC: #'Collections-Unordered'"
	"self default isDefinedAsPackageOrSubPackageInMC: #'Collections-Unordered-Test'"
	^ MCWorkingCopy allManagers 
			anySatisfy: [:each | self category: each packageName asSymbol matches: aSymbol ] ! !

!RPackageOrganizer methodsFor: 'mc synchronization' stamp: 'StephaneDucasse 7/29/2010 12:25'!
isDefinedPackageInMC: aSymbol
	"Answer whether a package is already defined with the exact name"
	"self default isDefinedAsSubPackageInMC: #Compression"
	"self default isDefinedAsSubPackageInMC: #'Compression-Archives'"
	^ MCWorkingCopy allManagers anySatisfy: [:each | each packageName asSymbol = aSymbol]! !

!RPackageOrganizer methodsFor: 'accessing' stamp: 'tg 3/13/2010 16:39'!
packageDefiningOrExtendingMethod: aCompiledMethod
	^ self packageDefiningOrExtendingSelector: aCompiledMethod selector inClass: aCompiledMethod methodClass
! !

!RPackageOrganizer methodsFor: 'accessing' stamp: 'tg 6/5/2010 11:38'!
packageDefiningOrExtendingSelector: aSelector inClass: aClass
	"this implementation is slower
		aClass packages detect: [:each | each includesSelector: aSelector ofClass: aClass ]"

	^ aClass packages detect: [:each | each includesSelector: aSelector ofClass: aClass ]
	
"	self packages detect: [:each | 
		each includesSelector: aSelector ofClass: aClass ].
"	
	! !

!RPackageOrganizer methodsFor: 'accessing' stamp: 'StephaneDucasse 7/29/2010 21:40'!
packageNamed: aSymbol

	^ packages at: aSymbol  ! !

!RPackageOrganizer methodsFor: 'accessing' stamp: 'stephane.ducasse 7/16/2009 23:17'!
packageNames
	^ packages keys! !

!RPackageOrganizer methodsFor: 'package access from class' stamp: 'tg 3/13/2010 16:28'!
packageOf: aClass

	self flag: #shouldProbablyReturnUnunpackagedPackage.
	^ classPackageMapping at: aClass theNonMetaClass name ifAbsent: [nil]! !

!RPackageOrganizer methodsFor: 'package access from class' stamp: 'cyrille.delaunay 11/24/2010 14:24'!
packageOfClassNamed: aName

	self flag: #shouldProbablyReturnUnunpackagedPackage.
	^ classPackageMapping at: aName asSymbol ifAbsent: [nil]! !

!RPackageOrganizer methodsFor: 'accessing' stamp: 'stephane.ducasse 7/10/2009 11:27'!
packages
	
	^ packages values! !

!RPackageOrganizer methodsFor: 'initialize' stamp: 'StephaneDucasse 7/16/2010 15:39'!
register
	MCWorkingCopy addDependent: self.
	self flag: 'to add listening to system announcer'.
! !

!RPackageOrganizer methodsFor: 'registration' stamp: 'sd 5/21/2010 14:44'!
registerExtendingPackage: aPackage forClass: aClass 

	| cur |
	cur := classExtendingPackagesMapping at: aClass theNonMetaClass name ifAbsent: [nil].
	cur isNil
		ifTrue: [classExtendingPackagesMapping at: aClass theNonMetaClass name put: (Set with: aPackage)]
		ifFalse: [cur add: aPackage]! !

!RPackageOrganizer methodsFor: 'registration' stamp: 'cyrille.delaunay 11/24/2010 14:06'!
registerExtendingPackage: aPackage forClassName: aClassName 

	| cur |
	cur := classExtendingPackagesMapping at: aClassName asSymbol ifAbsent: [nil].
	cur isNil
		ifTrue: [classExtendingPackagesMapping at: aClassName asSymbol put: (Set with: aPackage)]
		ifFalse: [cur add: aPackage]! !

!RPackageOrganizer methodsFor: 'registration' stamp: 'cyrille.delaunay 11/22/2010 13:29'!
registerPackage: aPackage
	"Transcript show: self class printString , ' registerPackage: ', aPackage packageName asString ; cr."
	^ packages at: aPackage name asSymbol put: aPackage.
	! !

!RPackageOrganizer methodsFor: 'registration' stamp: 'cyrille.delaunay 11/25/2010 15:08'!
registerPackage: aPackage forClass: aClass 
	(aPackage includesClass: aClass) 
		ifFalse: 
			[self error: aPackage name , ' does not includes the class ' , aClass name].
	^classPackageMapping at: aClass theNonMetaClass name put: aPackage! !

!RPackageOrganizer methodsFor: 'registration' stamp: 'cyrille.delaunay 11/25/2010 15:08'!
registerPackage: aPackage forClassName: aClassNameSymbol
	(aPackage includesClassNamed: aClassNameSymbol asSymbol) 
		ifFalse: 
			[self error: aPackage name , ' does not includes the class ' , aClassNameSymbol].
	^classPackageMapping at: aClassNameSymbol put: aPackage! !

!RPackageOrganizer methodsFor: 'registration' stamp: 'TudorGirba 11/26/2010 00:29'!
registerPackageNamed: aSymbol
	^ packages at: aSymbol ifAbsentPut: [RPackage named: aSymbol]! !

!RPackageOrganizer methodsFor: 'mc synchronization' stamp: 'StephaneDucasse 7/29/2010 20:32'!
shouldDefineAPackage: aSymbol
	"we should define a package if it is not created and it correspond to an MC package"
	"now this is not clear that we need that since an MC package is only created after creating a category"
	"Watch out!! Not tested"
	
	^ (self includesPackageNamed: aSymbol)
			ifTrue: [^false]
			ifFalse: [ (self isDefinedPackageInMC: aSymbol) 
						ifTrue: [^ true]
						ifFalse: [^ (self isDefinedAsPackageOrSubPackageInMC: aSymbol)]]! !

!RPackageOrganizer methodsFor: 'private' stamp: 'sd 5/21/2010 14:51'!
stopNotification
	"(self instVarNamed: #default) stopNotification"
	
	"pay attention that we can break the system using this method"
	
	SystemChangeNotifier uniqueInstance noMoreNotificationsFor: self.
	MCWorkingCopy removeDependent: self.! !

!RPackageOrganizer methodsFor: 'registration' stamp: 'sd 5/19/2010 16:13'!
unregisterExtendingPackage: aPackage forClass: aClass
	
	| cur |
	cur := classExtendingPackagesMapping at: aClass theNonMetaClass name ifAbsent: [nil].
	cur isNil
		ifFalse: [cur remove: aPackage ifAbsent: [] "not happy with this one"]! !

!RPackageOrganizer methodsFor: 'registration' stamp: 'cyrille.delaunay 11/24/2010 14:06'!
unregisterExtendingPackage: aPackage forClassName: aClassName
	
	| cur |
	cur := classExtendingPackagesMapping at: aClassName asSymbol ifAbsent: [nil].
	cur isNil
		ifFalse: [cur remove: aPackage ifAbsent: [] "not happy with this one"]! !

!RPackageOrganizer methodsFor: 'registration' stamp: 'sd 4/25/2010 15:51'!
unregisterPackage: aPackage
	"Transcript show: self class printString , ' unregisterPackage: ', aPackage packageName asString ; cr."
	
	packages removeKey: aPackage name asSymbol ifAbsent: [].	
	! !

!RPackageOrganizer methodsFor: 'registration' stamp: 'cyrille.delaunay 11/25/2010 15:39'!
unregisterPackage: aPackage forClass: aClass 
	^classPackageMapping removeKey: aClass theNonMetaClass name! !

!RPackageOrganizer methodsFor: 'registration' stamp: 'cyrille.delaunay 11/25/2010 15:39'!
unregisterPackage: aPackage forClassName: aClassName
	(aPackage includesClassNamed: aClassName asSymbol) 
		ifTrue: 
			[self error: aPackage name , ' still includes the class ' , aClassName asSymbol].
	^classPackageMapping removeKey: aClassName asSymbol! !

!RPackageOrganizer methodsFor: 'registration' stamp: 'StephaneDucasse 12/20/2009 11:22'!
unregisterPackageNamed: aSymbolName
	packages removeKey: aSymbolName asSymbol ifAbsent: [].	
	! !

!RPackageOrganizer methodsFor: 'mc synchronization' stamp: 'cyrille.delaunay 11/25/2010 15:53'!
update: anArray
	"User create a new package, MCWorkingCopy propagates changes. We react accordingly."
	
	super update: anArray.
	(anArray isArray and: [ anArray first = #registered ])
		ifTrue: 
			[ 
			| newPackage |
			"we first look if a RPackage has not yet been registered (by creating a category in the system). In this case we don't want to register a new RPackage, but refer to this existing category"
			"self haltIf: [anArray second name = 'Zork']."
			(self packageExactlyMatchingExtensionName: anArray second name) isNil
				ifTrue: [
					"if no package already exist, we add a new one"
					newPackage := self registerPackageNamed: anArray second name.
					newPackage importCategory.
					]
			"Probably can get a bug since we can have a package named AST and the category named AST-Core so should probably 
	pass the category as parameter"
			].
	(anArray isArray and: [ anArray first = #unregistered ])
		ifTrue: [ self unregisterPackageNamed: anArray second name.
				self flag: #shouldDoSomethingHere. ].! !

!PackageInfo methodsFor: '*RPackage-Core' stamp: 'cyrille.delaunay 11/25/2010 16:40'!
classesAndMetaClassesNotDefinedInSubCategories

	|subCategoriesClasses|
	subCategoriesClasses := OrderedCollection new.
	self systemCategories do: [:aSymbol |
		aSymbol = self packageName asSymbol
			ifFalse: [
				subCategoriesClasses addAll: (PackageOrganizer default packageNamed: aSymbol ifAbsent: [PackageInfo new]) classesAndMetaClasses
				]
		].
	^ self classesAndMetaClasses difference: subCategoriesClasses! !

!PackageInfo methodsFor: '*RPackage-Core' stamp: 'cyrille.delaunay 11/25/2010 16:40'!
classesNotDefinedInSubCategories

	|subCategoriesClasses|
	subCategoriesClasses := OrderedCollection new.
	self systemCategories do: [:aSymbol |
		aSymbol = self packageName asSymbol
			ifFalse: [
				subCategoriesClasses addAll: (PackageOrganizer default packageNamed: aSymbol ifAbsent: [PackageInfo new]) classes
				]
		].
	^ self classes difference: subCategoriesClasses! !

!PackageInfo methodsFor: '*RPackage-Core' stamp: 'cyrille.delaunay 10/29/2010 10:31'!
coreMethodsNotDefinedInSubCategories

	^ self classesAndMetaClassesNotDefinedInSubCategories gather: [:class | self coreMethodsForClass: class]! !

!PackageInfo methodsFor: '*RPackage-Core' stamp: 'cyrille.delaunay 10/29/2010 10:37'!
extensionMethodsNotDefinedInSubCategories
	^ self externalBehaviorsNotDefinedInSubCategories gather: [:classOrTrait | self extensionMethodsForClass: classOrTrait]! !

!PackageInfo methodsFor: '*RPackage-Core' stamp: 'cyrille.delaunay 10/29/2010 10:37'!
externalBehaviorsNotDefinedInSubCategories
	^self externalClassesNotDefinedInSubCategories , self externalTraitsNotDefinedInSubCategories! !

!PackageInfo methodsFor: '*RPackage-Core' stamp: 'cyrille.delaunay 10/29/2010 10:33'!
externalClassesNotDefinedInSubCategories
	| myClasses |
	myClasses := self classesAndMetaClassesNotDefinedInSubCategories.
	^ Array streamContents:
		[:s |
		ProtoObject withAllSubclassesDo:
			[:class |
			(myClasses includes: class) ifFalse: [s nextPut: class]]]! !

!PackageInfo methodsFor: '*RPackage-Core' stamp: 'cyrille.delaunay 10/29/2010 10:35'!
externalTraitsNotDefinedInSubCategories
	
	
	^ Array streamContents: [:s | | behaviors |
		behaviors := self classesAndMetaClassesNotDefinedInSubCategories.
		Smalltalk allTraits do: [:trait |
			(behaviors includes: trait) ifFalse: [s nextPut: trait].
			(behaviors includes: trait classSide) ifFalse: [s nextPut: trait classSide]]].			! !
RPackage initialize!
